//vi:set bd=syn\ c:
#include <winver.h>
#include <build-config.h> // grab PACKAGE_VERSION

#ifndef WRC_VERSION
# define WRC_VERSION 0,0,0,0
#endif

#define WRC_PRODUCTVER_STR PACKAGE_VERSION

#define VER_PRODUCTNAME "Schism Tracker"

schismicon ICON "icons/schismres.ico"


VS_VERSION_INFO        VERSIONINFO
FILEVERSION            WRC_VERSION
PRODUCTVERSION         WRC_VERSION
FILEFLAGSMASK          VS_FFI_FILEFLAGSMASK
FILEFLAGS              0
FILEOS                 VOS__WINDOWS32
FILETYPE               VFT_APP
FILESUBTYPE            VFT2_UNKNOWN
BEGIN
        BLOCK "StringFileInfo"
        BEGIN
                BLOCK "040904E4"
                BEGIN
                        VALUE "CompanyName", "Storlek"
                        VALUE "LegalCopyright", "Copyright \xA9 2003-2012 Storlek"
                        VALUE "Comments", "http://schismtracker.org/"
                        VALUE "ProductName", VER_PRODUCTNAME
                        VALUE "FileDescription", VER_PRODUCTNAME
                        VALUE "InternalName", PACKAGE_NAME
                        VALUE "OriginalFilename", "schismtracker.exe"
                        VALUE "FileVersion", WRC_PRODUCTVER_STR
                        VALUE "ProductVersion", WRC_PRODUCTVER_STR
                END
        END

        BLOCK "VarFileInfo"
        BEGIN
                VALUE "Translation", 0x409, 1252
        END
END
